"""Mysql module tests."""
from datetime import datetime
from unittest import mock

import pytest

from cumin import Config, NodeSet

from spicerack import mysql
from spicerack.remote import Remote, RemoteHosts

from spicerack.tests import caplog_not_available, get_fixture_path
from spicerack.tests.unit.test_remote import mock_cumin


class TestMysqlRemoteHosts:
    """Test class for the MysqlRemoteHosts class."""

    @mock.patch('spicerack.remote.transports', autospec=True)
    def setup_method(self, _, mocked_transports):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path('remote', 'config.yaml'))
        self.mocked_transports = mocked_transports
        self.mysql_remote_hosts = mysql.MysqlRemoteHosts(RemoteHosts(self.config, NodeSet('host[1-9]'), dry_run=False))
        self.expected = [(NodeSet('host1'), 'output1')]

    def test_execute(self):
        """Calling run_query() should run the given query in the target hosts."""
        mock_cumin(self.mocked_transports, 0)
        results = self.mysql_remote_hosts.run_query('query1')
        assert [(host, msg.message().decode()) for host, msg in results] == self.expected


class TestMysql:
    """Mysql class tests."""

    @mock.patch('spicerack.remote.transports', autospec=True)
    def setup_method(self, _, mocked_transports):
        """Initialize the test environment for Mysql."""
        # pylint: disable=attribute-defined-outside-init
        self.config = Config(get_fixture_path('remote', 'config.yaml'))
        self.mocked_transports = mocked_transports
        self.mocked_remote = mock.MagicMock(spec_set=Remote)
        self.mysql = mysql.Mysql(self.mocked_remote, dry_run=False)

    def test_get_dbs(self):
        """It should return and instance of MysqlRemoteHosts for the matching hosts."""
        self.mysql.get_dbs('query')
        self.mocked_remote.query.assert_called_once_with('query')

    @pytest.mark.parametrize('kwargs, query, match', (
        ({}, 'P{O:mariadb::core}', 'db10[01-99],db20[01-99]'),
        ({'datacenter': 'eqiad'}, 'P{O:mariadb::core} and A:eqiad', 'db10[01-99]'),
        ({'section': 's1'}, 'P{O:mariadb::core} and P{C:mariadb::heartbeat and R:Class%shard = "s1"}',
         'db10[01-10],db20[01-10]'),
        ({'replication_role': 'master'},
         'P{O:mariadb::core} and P{C:mariadb::config and R:Class%replication_role = "master"}',
         'db10[01-11],db20[01-11]'),
        ({'datacenter': 'eqiad', 'section': 's1'},
         'P{O:mariadb::core} and A:eqiad and P{C:mariadb::heartbeat and R:Class%shard = "s1"}', 'db10[01-10]'),
        ({'datacenter': 'eqiad', 'replication_role': 'master'},
         'P{O:mariadb::core} and A:eqiad and P{C:mariadb::config and R:Class%replication_role = "master"}',
         'db10[01-11]'),
        ({'section': 's1', 'replication_role': 'master'},
         ('P{O:mariadb::core} and P{C:mariadb::heartbeat and R:Class%shard = "s1"} and '
          'P{C:mariadb::config and R:Class%replication_role = "master"}'), 'db1001,db2001'),
        ({'datacenter': 'eqiad', 'section': 's1', 'replication_role': 'master'},
         ('P{O:mariadb::core} and A:eqiad and P{C:mariadb::heartbeat and R:Class%shard = "s1"} and '
          'P{C:mariadb::config and R:Class%replication_role = "master"}'), 'db1001'),
    ))
    def test_get_core_dbs_ok(self, kwargs, query, match):
        """It should return the right DBs based on the parameters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet(match))
        self.mysql.get_core_dbs(**kwargs)
        self.mocked_remote.query.assert_called_once_with(query)

    @pytest.mark.parametrize('kwargs', (
        {'datacenter': 'invalid'},
        {'section': 'invalid'},
        {'replication_role': 'invalid'},
    ))
    def test_get_core_dbs_fail(self, kwargs):
        """It should raise MysqlError if called with invalid parameters."""
        message = 'Got invalid {key}'.format(key=list(kwargs.keys())[0])
        with pytest.raises(mysql.MysqlError, match=message):
            self.mysql.get_core_dbs(**kwargs)

        assert not self.mocked_remote.query.called

    def test_get_core_dbs_fail_sanity_check(self):
        """It should raise MysqlError if matching an invalid number of hosts when looking for masters."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        with pytest.raises(mysql.MysqlError, match='Matched 1 masters, expected 11'):
            self.mysql.get_core_dbs(datacenter='eqiad', replication_role='master')

        assert self.mocked_remote.query.called

    @pytest.mark.skipif(caplog_not_available(), reason='Requires caplog fixture')
    @pytest.mark.parametrize('mode, value', (('readonly', b'1'), ('readwrite', b'0')))
    def test_set_core_masters_readonly(self, mode, value, caplog):
        """It should set the masters as read-only/read-write."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db10[01-11]'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db10[01-11]', value)]])
        getattr(self.mysql, 'set_core_masters_' + mode)('eqiad')
        assert 'SET GLOBAL read_only=' + value.decode() in caplog.text

    @pytest.mark.skipif(caplog_not_available(), reason='Requires caplog fixture')
    @pytest.mark.parametrize('readonly, reply', ((True, b'1'), (False, b'0')))
    def test_verify_core_masters_readonly_ok(self, readonly, reply, caplog):
        """Should verify that the masters have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db10[01-11]'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db10[01-11]', reply)]])
        self.mysql.verify_core_masters_readonly('eqiad', readonly)
        assert 'SELECT @@global.read_only' in caplog.text

    def test_verify_core_masters_readonly_fail(self):
        """Should raise MysqlError if some masters do not have the intended read-only value."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db10[01-11]'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db1001', b'0'), ('db10[02-11]', b'1')]])
        with pytest.raises(mysql.MysqlError, match='Verification failed that core DB masters'):
            self.mysql.verify_core_masters_readonly('eqiad', True)

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_in_sync_ok(self, mocked_sleep):
        """Should check that all core masters are in sync with the master in the other DC."""
        hosts = NodeSet('db10[01-11]')
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, NodeSet(host)) for host in hosts] * 2
        retvals = [[(host, b'2018-09-06T10:00:00.000000')] for host in hosts]  # first heartbeat
        retvals += [[(host, b'2018-09-06T10:00:01.000000')] for host in hosts]  # second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        self.mysql.check_core_masters_in_sync('eqiad', 'codfw')
        assert not mocked_sleep.called

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_in_sync_fail_heartbeat(self, mocked_sleep):
        """Should raise MysqlError if unable to get the heartbeat from the current master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql.MysqlError, match='Unable to get heartbeat from master'):
            self.mysql.check_core_masters_in_sync('eqiad', 'codfw')
        assert not mocked_sleep.called

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_in_sync_not_in_sync(self, mocked_sleep):
        """Should raise MysqlError if a master is not in sync with the one in the other DC."""
        hosts = NodeSet('db10[01-11]')
        self.mocked_remote.query.side_effect = [RemoteHosts(self.config, NodeSet(host)) for host in hosts] + [
            RemoteHosts(self.config, NodeSet('db1001'))] * 3
        retvals = [[(host, b'2018-09-06T10:00:00.000000')] for host in hosts]  # first heartbeat
        retvals += [[('db1001', b'2018-09-06T10:00:00.000000')]] * 3  # 3 failed retries of second heartbeat
        mock_cumin(self.mocked_transports, 0, retvals=retvals)
        with pytest.raises(mysql.MysqlError, match=r'Heartbeat from master db1001 for section .* not yet in sync'):
            self.mysql.check_core_masters_in_sync('eqiad', 'codfw')

        assert mocked_sleep.called

    def test_get_core_masters_heartbeats_wrong_data(self):
        """Should raise MysqlError if unable to convert the heartbeat into a datetime."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        mock_cumin(self.mocked_transports, 0, retvals=[[('db1001', b'2018-09-06-10:00:00.000000')]])
        with pytest.raises(mysql.MysqlError, match='Unable to convert heartbeat'):
            self.mysql.get_core_masters_heartbeats('eqiad', 'codfw')

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_check_core_masters_heartbeats_fail(self, mocked_sleep):
        """Should raise MysqlError if unable to get the heartbeat from the master."""
        self.mocked_remote.query.return_value = RemoteHosts(self.config, NodeSet('db1001'))
        mock_cumin(self.mocked_transports, 0, retvals=[])
        with pytest.raises(mysql.MysqlError, match='Unable to get heartbeat from master'):
            self.mysql.check_core_masters_heartbeats('eqiad', 'codfw', {'s1': datetime.utcnow()})

        assert mocked_sleep.called



"""ElasticsearchCluster module test."""
from datetime import datetime, timedelta
from unittest import mock

import pytest

from cumin import NodeSet
from elasticsearch import Elasticsearch, ConflictError, RequestError, TransportError

from spicerack import elasticsearch_cluster as ec
from spicerack.administrative import Reason
from spicerack.remote import Remote, RemoteHosts
from spicerack.tests import elasticsearch_too_old


pytestmark = pytest.mark.skipif(  # pylint: disable=invalid-name
    elasticsearch_too_old(), reason='Requires more recent elasticsearch module'
)


def test_create_elasticsearch_clusters():
    """It should return an instance of ElasticsearchCluster."""
    target = ec.create_elasticsearch_clusters('search_eqiad', None)
    assert isinstance(target, ec.ElasticsearchClusters)


def test_create_elasticsearch_clusters_fail():
    """It should throw an ElasticsearchCluster Exception."""
    with pytest.raises(ec.ElasticsearchClusterError, match='No cluster group named search_test'):
        ec.create_elasticsearch_clusters('search_test', None)


def test_get_remote_hosts():
    """Test that RemoteHosts instance is returned."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.hosts = NodeSet('el[1-2]')
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    result = elastic_hosts.get_remote_hosts()
    assert isinstance(result, RemoteHosts)


def test_start_elasticsearch():
    """Test that start elasticsearch service is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.start_elasticsearch()
    mocked_remote_hosts.run_sync.assert_called_with('cat /etc/elasticsearch/instances | xarg systemctl start')


def test_stop_elasticsearch():
    """Test that stop elasticsearch service is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.stop_elasticsearch()
    mocked_remote_hosts.run_sync.assert_called_with('cat /etc/elasticsearch/instances | xarg systemctl stop')


def test_restart_elasticsearch():
    """Test that restart elasticsearch service is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.restart_elasticsearch()
    mocked_remote_hosts.run_sync.assert_called_with('cat /etc/elasticsearch/instances | xarg systemctl restart')


def test_depool_nodes():
    """Test that depool command is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.depool_nodes()
    mocked_remote_hosts.run_sync.assert_called_with('depool')


def test_pool_nodes():
    """Test that pool command is called correctly."""
    mocked_remote_hosts = mock.Mock(spec_set=RemoteHosts)
    mocked_remote_hosts.run_sync = mock.Mock(return_value=iter(()))
    elastic_hosts = ec.ElasticsearchHosts(mocked_remote_hosts, None)
    elastic_hosts.pool_nodes()
    mocked_remote_hosts.run_sync.assert_called_with('pool')


def test_elasticsearch_call_when_wait_for_elasticsearch_up():
    """Test that elasticsearch instance is called when wait for instance to come up."""
    remote = mock.Mock(spec_set=Remote)
    elastic_alpha = Elasticsearch('localhost:9200')
    elastic_alpha.nodes.info = mock.Mock(return_value={
        'nodes': {
            'ELASTIC4': {'name': 'el5-alpha'}, 'ELASTIC5': {'name': 'el6-alpha'}
        }
    })
    elastic_alpha_cluster = ec.ElasticsearchCluster(elastic_alpha, None, dry_run=False)

    elastic_beta = Elasticsearch('localhost:9201')
    elastic_beta.nodes.info = mock.Mock(return_value={
        'nodes': {
            'ELASTIC4': {'name': 'el5-beta'}, 'ELASTIC5': {'name': 'el6-beta'}
        }
    })
    elastic_beta_cluster = ec.ElasticsearchCluster(elastic_beta, None, dry_run=False)

    next_nodes = [
        {'name': 'el5',
            'clusters_instances': [
                elastic_alpha_cluster,
                elastic_beta_cluster,
            ]
         },
        {'name': 'el6',
            'clusters_instances': [
                elastic_alpha_cluster,
                elastic_beta_cluster,
            ]
         }
    ]

    elastic_hosts = ec.ElasticsearchHosts(remote, next_nodes, dry_run=False)
    elastic_hosts.wait_for_elasticsearch_up()
    assert elastic_alpha.nodes.info.called
    assert elastic_beta.nodes.info.called


def test_elasticsearch_call_not_made_when_wait_for_elasticsearch_up_in_dry_run():
    """Test that elasticsearch instance is not called in dry run mode."""
    remote = mock.Mock(spec_set=Remote)
    elastic_alpha = Elasticsearch('localhost:9200')
    elastic_alpha.nodes.info = mock.Mock(return_value={
        'nodes': {
            'ELASTIC4': {'name': 'el5-alpha'}, 'ELASTIC5': {'name': 'el6-alpha'}
        }
    })
    elastic_alpha_cluster = ec.ElasticsearchCluster(elastic_alpha, None, dry_run=False)

    elastic_beta = Elasticsearch('localhost:9201')
    elastic_beta.nodes.info = mock.Mock(return_value={
        'nodes': {
            'ELASTIC4': {'name': 'el5-beta'}, 'ELASTIC5': {'name': 'el6-beta'}
        }
    })
    elastic_beta_cluster = ec.ElasticsearchCluster(elastic_beta, None, dry_run=False)

    next_nodes = [
        {'name': 'el5',
         'clusters_instances': [
             elastic_alpha_cluster,
             elastic_beta_cluster,
         ]
         },
        {'name': 'el6',
         'clusters_instances': [
             elastic_alpha_cluster,
             elastic_beta_cluster,
         ]
         }
    ]

    elastic_hosts = ec.ElasticsearchHosts(remote, next_nodes, dry_run=True)
    elastic_hosts.wait_for_elasticsearch_up()
    assert not elastic_alpha.nodes.info.called
    assert not elastic_beta.nodes.info.called


@mock.patch('spicerack.decorators.time.sleep', return_value=None)
def test_elasticsearch_call_is_retried_when_wait_for_elasticsearch_up_is_not_up(mocked_sleep):
    """Test that elasticsearch instance is called more than once when node is not found in cluster."""
    remote = mock.Mock(spec_set=Remote)
    elastic_alpha = Elasticsearch('localhost:9200')
    elastic_alpha.nodes.info = mock.Mock(return_value={
        'nodes': {
            'ELASTIC4': {'name': 'el7-alpha'}, 'ELASTIC5': {'name': 'el8-alpha'}
        }
    })
    elastic_alpha_cluster = ec.ElasticsearchCluster(elastic_alpha, None, dry_run=False)

    elastic_beta = Elasticsearch('localhost:9201')
    elastic_beta.nodes.info = mock.Mock(return_value={
        'nodes': {
            'ELASTIC4': {'name': 'el5-beta'}, 'ELASTIC5': {'name': 'el6-beta'}
        }
    })
    elastic_beta_cluster = ec.ElasticsearchCluster(elastic_beta, None, dry_run=False)

    next_nodes = [
        {'name': 'el5',
         'clusters_instances': [
             elastic_alpha_cluster,
             elastic_beta_cluster,
         ]
         },
        {'name': 'el6',
         'clusters_instances': [
             elastic_alpha_cluster,
             elastic_beta_cluster,
         ]
         }
    ]

    elastic_hosts = ec.ElasticsearchHosts(remote, next_nodes, dry_run=False)
    with pytest.raises(ec.ElasticsearchClusterCheckError):
        elastic_hosts.wait_for_elasticsearch_up(timedelta(seconds=20))
        assert mocked_sleep.called
    assert elastic_alpha.nodes.info.call_count == 4


@mock.patch('spicerack.decorators.time.sleep', return_value=None)
def test_elasticsearch_call_is_retried_when_wait_for_elasticsearch_up_cannot_be_reached(mocked_sleep):
    """Test that elasticsearch instance is called more than once when the cluster cannot be reached."""
    remote = mock.Mock(spec_set=Remote)
    elastic_alpha = Elasticsearch('localhost:9200')
    elastic_alpha.nodes.info = mock.Mock(side_effect=TransportError)
    elastic_alpha_cluster = ec.ElasticsearchCluster(elastic_alpha, None, dry_run=False)

    next_nodes = [{
        'name': 'el5',
        'clusters_instances': [elastic_alpha_cluster]
    }]

    elastic_hosts = ec.ElasticsearchHosts(remote, next_nodes, dry_run=False)
    with pytest.raises(ec.ElasticsearchClusterError):
        elastic_hosts.wait_for_elasticsearch_up(timedelta(seconds=20))
        assert mocked_sleep.called
    assert elastic_alpha.nodes.info.call_count == 4


def test_cluster_settings_are_unchanged_when_stopped_replication_is_dry_run():
    """Check that cluster routing in dry run mode is truly safe"""
    elasticsearch = Elasticsearch('endpoint:9200')
    elasticsearch.cluster.put_settings = mock.Mock(return_value=True)
    elasticsearchcluster = ec.ElasticsearchCluster(elasticsearch, None, dry_run=True)
    with elasticsearchcluster.stopped_replication():
        assert not elasticsearch.cluster.put_settings.called


def test_split_node_names():
    """split_node_name() support cluster names containing '-'"""
    node_name, cluster_name = ec.ElasticsearchCluster.split_node_name('node1-cluster-name')
    assert node_name == 'node1'
    assert cluster_name == 'cluster-name'


class TestElasticsearchClusters:
    """Test class for Elasticsearch Clusters"""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.elasticsearch1 = Elasticsearch('endpoint:9201')
        self.elasticsearch2 = Elasticsearch('endpoint:9202')
        self.cluster1 = ec.ElasticsearchCluster(self.elasticsearch1, None, dry_run=False)
        self.cluster2 = ec.ElasticsearchCluster(self.elasticsearch2, None, dry_run=False)
        self.clusters = [self.cluster1, self.cluster2]

    def test_flush_markers_on_clusters(self):
        """Test that elasticsearch call to flush markers was properly made."""
        self.elasticsearch1.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch1.indices.flush_synced = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush_synced = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.flush_markers(timedelta(seconds=30))
        self.elasticsearch1.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch1.indices.flush_synced.assert_called_with(request_timeout=30)
        self.elasticsearch2.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch2.indices.flush_synced.assert_called_with(request_timeout=30)

    def test_flush_markers_on_clusters_fail(self):
        """Test that when conflict error is raised, execution continues."""
        self.elasticsearch1.indices.flush = mock.Mock(side_effect=ConflictError('test'))
        self.elasticsearch1.indices.flush_synced = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush_synced = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.flush_markers(timedelta(seconds=30))
        self.elasticsearch1.indices.flush_synced.assert_called_with(request_timeout=30)
        self.elasticsearch2.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch2.indices.flush_synced.assert_called_with(request_timeout=30)

    def test_flush_markers_on_clusters_fail_synced(self):
        """Test that when conflict error is raised during synced flush, execution continues."""
        self.elasticsearch1.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch1.indices.flush_synced = mock.Mock(side_effect=ConflictError('test'))
        self.elasticsearch2.indices.flush = mock.Mock(return_value=True)
        self.elasticsearch2.indices.flush_synced = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.flush_markers(timedelta(seconds=30))
        self.elasticsearch1.indices.flush_synced.assert_called_with(request_timeout=30)
        self.elasticsearch2.indices.flush.assert_called_with(force=True, request_timeout=30)
        self.elasticsearch2.indices.flush_synced.assert_called_with(request_timeout=30)

    def test_when_all_shards_are_assigned_no_allocation_is_performed(self):
        """Test that shard allocation is not performed when all shards have been assigned on all clusters"""
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': {'ELASTIC1': {'name': 'el1-alpha'}}})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': {'ELASTIC7': {'name': 'el1-beta'}}})
        self.elasticsearch1.cat.shards = mock.Mock(return_value=[
            {'index': 'index1', 'shard': 2, 'state': 'ASSIGNED'},
            {'index': 'index2', 'shard': 4, 'state': 'ASSIGNED'}
        ])
        self.elasticsearch2.cat.shards = mock.Mock(return_value=[
            {'index': 'index3', 'shard': 6, 'state': 'ASSIGNED'},
            {'index': 'index4', 'shard': 7, 'state': 'ASSIGNED'}
        ])
        self.elasticsearch1.cluster.reroute = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.reroute = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.force_allocation_of_all_unassigned_shards()
        assert not self.elasticsearch1.cluster.reroute.called
        assert not self.elasticsearch2.cluster.reroute.called

    def test_force_allocation_of_all_unassigned_shards(self):
        """Test that elasticsearch performs cluster reroute with with unassigned shards on all clusters."""
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': {'ELASTIC1': {'name': 'el1-alpha'}}})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': {'ELASTIC7': {'name': 'el1-beta'}}})
        self.elasticsearch1.cat.shards = mock.Mock(return_value=[
            {'index': 'index1', 'shard': 2, 'state': 'UNASSIGNED'},
            {'index': 'index2', 'shard': 4, 'state': 'ASSIGNED'}
        ])
        self.elasticsearch2.cat.shards = mock.Mock(return_value=[
            {'index': 'index3', 'shard': 6, 'state': 'ASSIGNED'},
            {'index': 'index4', 'shard': 7, 'state': 'UNASSIGNED'}
        ])
        self.elasticsearch1.cluster.reroute = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.reroute = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.force_allocation_of_all_unassigned_shards()
        self.elasticsearch1.cluster.reroute.assert_called_with(retry_failed=True, body={
            'commands': [{
                'allocate_replica': {
                    'index': 'index1', 'shard': 2,
                    'node': 'el1-alpha'
                }
            }]
        })
        self.elasticsearch2.cluster.reroute.assert_called_with(retry_failed=True, body={
            'commands': [{
                'allocate_replica': {
                    'index': 'index4', 'shard': 7,
                    'node': 'el1-beta'
                }
            }]
        })

    def test_force_allocation_of_shards_with_failed_node(self):
        """Test that shard allocation command was called twice after it fails on the first node."""
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': {
            'ELASTIC1': {'name': 'el1-alpha'}, 'ELASTIC3': {'name': 'el2-alpha'}
        }})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': {
            'ELASTIC7': {'name': 'el1-beta'}, 'ELASTIC4': {'name': 'el2-beta'},
        }})
        self.elasticsearch1.cat.shards = mock.Mock(return_value=[
            {'index': 'index1', 'shard': 2, 'state': 'UNASSIGNED'},
            {'index': 'index2', 'shard': 4, 'state': 'ASSIGNED'}
        ])
        self.elasticsearch2.cat.shards = mock.Mock(return_value=[
            {'index': 'index3', 'shard': 6, 'state': 'ASSIGNED'},
            {'index': 'index4', 'shard': 7, 'state': 'UNASSIGNED'}
        ])
        self.elasticsearch1.cluster.reroute = mock.Mock(side_effect=RequestError('test'))
        self.elasticsearch2.cluster.reroute = mock.Mock(side_effect=RequestError('test'))
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.force_allocation_of_all_unassigned_shards()
        assert self.elasticsearch1.cluster.reroute.call_count == 2
        assert self.elasticsearch2.cluster.reroute.call_count == 2

    def test_stopped_replication(self):
        """Check that context manager stops replication and then starts replication on each cluster."""
        self.elasticsearch1.cluster.put_settings = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.put_settings = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        with elasticsearchclusters.stopped_replication():
            self.elasticsearch1.cluster.put_settings.assert_called_with(body={
                'transient': {
                    'cluster.routing.allocation.enable': 'primaries'
                }
            })
            self.elasticsearch2.cluster.put_settings.assert_called_with(body={
                'transient': {
                    'cluster.routing.allocation.enable': 'primaries'
                }
            })

        self.elasticsearch1.cluster.put_settings.assert_called_with(body={
            'transient': {
                'cluster.routing.allocation.enable': 'all'
            }
        })
        self.elasticsearch2.cluster.put_settings.assert_called_with(body={
            'transient': {
                'cluster.routing.allocation.enable': 'all'
            }
        })

    def test_frozen_writes_write_to_index(self):
        """Test that elasticsearch write to index is called to freeze writes"""
        self.elasticsearch1.index = mock.Mock(return_value=True)
        self.elasticsearch2.index = mock.Mock(return_value=True)
        self.elasticsearch1.delete = mock.Mock(return_value=True)
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        reason = Reason('test', 'test_user', 'test_host', task_id='T111222')
        cluster1 = ec.ElasticsearchCluster(self.elasticsearch1, None, dry_run=False)
        cluster2 = ec.ElasticsearchCluster(self.elasticsearch2, None, dry_run=False)
        elasticsearchclusters = ec.ElasticsearchClusters([cluster1, cluster2], None)
        with elasticsearchclusters.frozen_writes(reason):
            assert self.elasticsearch1.index.called
            assert self.elasticsearch2.index.called

        assert self.elasticsearch1.delete.called
        assert self.elasticsearch2.delete.called

    def test_when_frozen_writes_fails_exception_is_raised(self):
        """Test that when elasticsearch write to index fails, an exception is raised

        and a call to delete/unfreeze write is placed
        """
        self.elasticsearch1.index = mock.Mock(side_effect=TransportError('test'))
        self.elasticsearch2.index = mock.Mock(return_value=True)
        self.elasticsearch1.delete = mock.Mock(return_value=True)
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        reason = Reason('test', 'test_user', 'test_host', task_id='T111222')
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        with pytest.raises(ec.ElasticsearchClusterError):
            with elasticsearchclusters.frozen_writes(reason):
                assert self.elasticsearch1.index.called

            assert self.elasticsearch1.delete.called
            assert self.elasticsearch2.delete.called

    def test_when_unfreeze_writes_fails_exception_is_raised(self):
        """Test that when elasticsearch delete doc in index fails, an exception is raised

        and a call to delete/unfreeze write is placed
        """
        self.elasticsearch1.index = mock.Mock(return_value=True)
        self.elasticsearch2.index = mock.Mock(return_value=True)
        self.elasticsearch1.delete = mock.Mock(side_effect=TransportError('test'))
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        reason = Reason('test', 'test_user', 'test_host', task_id='T111222')
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        with pytest.raises(ec.ElasticsearchClusterError):
            with elasticsearchclusters.frozen_writes(reason):
                assert self.elasticsearch1.index.called

            assert self.elasticsearch1.delete.called
            assert self.elasticsearch2.delete.called

    def test_no_call_to_freeze_write_in_dry_run(self):
        """Test that when dry run is enabled, call to write to cluster index to freeze write is not placed"""
        self.elasticsearch1.index = mock.Mock(return_value=True)
        self.elasticsearch2.delete = mock.Mock(return_value=True)
        cluster1 = ec.ElasticsearchCluster(self.elasticsearch1, None, dry_run=True)
        cluster2 = ec.ElasticsearchCluster(self.elasticsearch2, None, dry_run=True)
        reason = Reason('test', 'test_user', 'test_host', task_id='T111222')
        elasticsearchclusters = ec.ElasticsearchClusters([cluster1, cluster2], None)
        with elasticsearchclusters.frozen_writes(reason):
            assert not self.elasticsearch1.index.called
            assert not self.elasticsearch2.delete.called

    def test_wait_for_green_on_all_clusters_elastisearch_call(self):
        """Makes sure the call to elasticsearch.cluster.health is placed for each cluster."""
        self.elasticsearch1.cluster.health = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.health = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.wait_for_green(timedelta(seconds=13))
        assert self.elasticsearch1.cluster.health.called
        assert self.elasticsearch2.cluster.health.called

    @mock.patch('spicerack.elasticsearch_cluster.retry')
    def test_wait_for_green_correct_tries_test(self, retry):
        """Check that the number of tries is correctly computed."""
        self.elasticsearch1.cluster.health = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.health = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.wait_for_green(timedelta(seconds=20))
        assert retry.call_args[1]['tries'] == 2

    @mock.patch('spicerack.elasticsearch_cluster.retry')
    def test_wait_for_green_default_tries_test(self, retry):
        """Checks that a default value of 1 is returned when timeout is less than 10"""
        self.elasticsearch1.cluster.health = mock.Mock(return_value=True)
        self.elasticsearch2.cluster.health = mock.Mock(return_value=True)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        elasticsearchclusters.wait_for_green(timedelta(seconds=4))
        assert retry.call_args[1]['tries'] == 1

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    def test_wait_for_green_retry_test(self, mocked_sleep):
        """Test that the retry is called again when cluster health request throws an exception."""
        self.elasticsearch1.cluster.health = mock.Mock(side_effect=TransportError('test'))
        self.elasticsearch2.cluster.health = mock.Mock(side_effect=TransportError('test'))
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        with pytest.raises(ec.ElasticsearchClusterCheckError):
            elasticsearchclusters.wait_for_green(timedelta(seconds=20))
            assert mocked_sleep.called
            assert self.elasticsearch1.cluster.health.call_count == 2
            assert self.elasticsearch2.cluster.health.call_count == 2

    def test_get_next_clusters_nodes(self):
        """Test that next nodes belong in the same row on each cluster."""
        remote = mock.Mock(spec_set=Remote)
        since = datetime.utcfromtimestamp(20 / 1000)
        cluster1_nodes = {
            'ELASTIC1':
                {'name': 'el1-alpha', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 10}},
            'ELASTIC2':
                {'name': 'el2-alpha', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 10}},
            'ELASTIC3':
                {'name': 'el3-alpha', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 30}},
        }

        cluster2_nodes = {
            'ELASTIC4':
                {'name': 'el3-gamma', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 10}},
            'ELASTIC5':
                {'name': 'el4-gamma', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 10}},
        }
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': cluster1_nodes})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': cluster2_nodes})
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, remote, dry_run=False)
        elasticsearchclusters.get_next_clusters_nodes(since, 4)
        nodes_not_restarted = remote.query.call_args[0][0]
        nodes_not_restarted = nodes_not_restarted.split(',')
        nodes_not_restarted.sort()
        assert nodes_not_restarted == ['el1*', 'el2*', 'el3*', 'el4*']

    def test_get_next_clusters_nodes_raises_error_when_size_is_less_than_one(self):
        """Test that next nodes belong in the same row on each cluster."""
        since = datetime.utcfromtimestamp(20 / 1000)
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None, dry_run=False)
        with pytest.raises(ec.ElasticsearchClusterError):
            elasticsearchclusters.get_next_clusters_nodes(since, 0)

    def test_get_next_nodes_returns_less_nodes_than_specified(self):
        """Test that the nodes returned is less than specified based on if they have been restarted for each clusters"""
        remote = mock.Mock(spec_set=Remote)
        since = datetime.utcfromtimestamp(20 / 1000)
        cluster1_nodes = {
            'ELASTIC3':
                {'name': 'el3-alpha', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 30}},
            'ELASTIC4':
                {'name': 'el5-alpha', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 30}},
            'ELASTIC5':
                {'name': 'el6-alpha', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 10}},
        }

        cluster2_nodes = {
            'ELASTIC6':
                {'name': 'el3-beta', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 30}},
            'ELASTIC7':
                {'name': 'el5-beta', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 30}},
            'ELASTIC8':
                {'name': 'el7-beta', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 10}},
        }
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': cluster1_nodes})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': cluster2_nodes})
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, remote, dry_run=False)
        elasticsearchclusters.get_next_clusters_nodes(since, 4)
        nodes_not_restarted = remote.query.call_args[0][0]
        nodes_not_restarted = nodes_not_restarted.split(',')
        nodes_not_restarted.sort()
        assert nodes_not_restarted == ['el6*', 'el7*']

    def test_get_next_nodes_most_not_restarted(self):
        """Test to get rows that have the most not restarted nodes first on each cluster."""
        remote = mock.Mock(spec_set=Remote)
        since = datetime.utcfromtimestamp(20 / 1000)
        cluster1_nodes = {
            'ELASTIC4':
                {'name': 'el5-alpha', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 10}},
            'ELASTIC5':
                {'name': 'el6-alpha', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 10}},
            'ELASTIC6':
                {'name': 'el7-alpha', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 30}},
        }

        cluster2_nodes = {
            'ELASTIC7':
                {'name': 'el9-beta', 'attributes': {'row': 'row3'}, 'jvm': {'start_time_in_millis': 10}},
            'ELASTIC8':
                {'name': 'el8-beta', 'attributes': {'row': 'row3'}, 'jvm': {'start_time_in_millis': 50}},
            'ELASTIC9':
                {'name': 'el10-beta', 'attributes': {'row': 'row3'}, 'jvm': {'start_time_in_millis': 30}},
        }
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': cluster1_nodes})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': cluster2_nodes})
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, remote, dry_run=False)
        elasticsearchclusters.get_next_clusters_nodes(since, 4)
        nodes_not_restarted = remote.query.call_args[0][0]
        nodes_not_restarted = nodes_not_restarted.split(',')
        nodes_not_restarted.sort()
        assert nodes_not_restarted == ['el5*', 'el6*']

    def test_get_next_nodes_no_rows(self):
        """Test that all nodes have been restarted on all clusters."""
        since = datetime.utcfromtimestamp(20 / 1000)
        cluster1_nodes = {
            'ELASTIC3':
                {'name': 'el3-gamma', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 30}},
            'ELASTIC4':
                {'name': 'el5-gamma', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 87}},
            'ELASTIC5':
                {'name': 'el6-gamma', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 77}},
        }

        cluster2_nodes = {
            'ELASTIC3':
                {'name': 'el4-alpha', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 40}},
            'ELASTIC4':
                {'name': 'el5-gamma', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 89}},
            'ELASTIC5':
                {'name': 'el6-alpha', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 79}},
        }
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': cluster1_nodes})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': cluster2_nodes})
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        result = elasticsearchclusters.get_next_clusters_nodes(since, 2)
        assert result is None

    def test_get_next_nodes_fails_when_rows_are_not_same(self):
        """Test that error is raised when clusters instances of the same node belong to different rows"""
        since = datetime.utcfromtimestamp(20 / 1000)
        cluster1_nodes = {
            'ELASTIC3':
                {'name': 'el3-gamma', 'attributes': {'row': 'row1'}, 'jvm': {'start_time_in_millis': 10}},
            'ELASTIC4':
                {'name': 'el5-gamma', 'attributes': {'row': 'row2'}, 'jvm': {'start_time_in_millis': 87}},
        }

        cluster2_nodes = {
            'ELASTIC3':
                {'name': 'el3-alpha', 'attributes': {'row': 'row6'}, 'jvm': {'start_time_in_millis': 10}},

        }
        self.elasticsearch1.nodes.info = mock.Mock(return_value={'nodes': cluster1_nodes})
        self.elasticsearch2.nodes.info = mock.Mock(return_value={'nodes': cluster2_nodes})
        elasticsearchclusters = ec.ElasticsearchClusters(self.clusters, None)
        with pytest.raises(ec.ElasticsearchClusterError):
            elasticsearchclusters.get_next_clusters_nodes(since, 2)



"""Top level cookbook"""
import argparse


def get_title(args):
    """Calculate the title based on the args."""
    return '{doc}: {args}'.format(doc=__doc__, args=args)


def argument_parser():
    """As required by spicerack.cookbook."""
    return argparse.ArgumentParser('Argparse')


def run(args, _):
    """As required by spicerack.cookbook."""
    print(args)
    return 0



"""Confctl module to abstract Conftool."""
import logging
import re

from conftool import configuration, kvobject, loader
from conftool.drivers import BackendError

from spicerack.exceptions import SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ConfctlError(SpicerackError):
    """Custom exception class for errors of this module."""


class Confctl:
    """Confctl class to abstract conftool operations."""

    def __init__(self, config='/etc/conftool/config.yaml', schema='/etc/conftool/schema.yaml', dry_run=True):
        """Initialize the instance.

        Arguments:
            config (str, optional): the path to the configuration file to load.
            schema (str, optional): the path to the Conftool schema to load.
            dry_run (bool, optional): whether this is a DRY-RUN.
        """
        self._dry_run = dry_run
        self._schema = loader.Schema.from_file(schema)
        kvobject.KVObject.setup(configuration.get(config))

    def entity(self, entity_name):
        """Get the Conftool specific entity class.

        Arguments:
            entity_name (str): the name of the entiryself.

        Returns:
            spicerack.confctl.ConftoolEntity: and entity-specific class to perform Conftool operations.

        """
        return ConftoolEntity(self._schema.entities[entity_name], dry_run=self._dry_run)


class ConftoolEntity:
    """ConftoolEntity class to perform operations on a specific Conftool entity."""

    def __init__(self, entity, dry_run=True):
        """Initialize the instance.

        Arguments:
            entity (conftool.kvobject.Entity): an instance of Conftool entity.
            dry_run (bool, optional): whether this is a DRY-RUN.
        """
        self._entity = entity
        self._dry_run = dry_run

    def _select(self, tags):
        """Generator that yields the selected objects based on the provided tags.

        Arguments:
            tags (dict): dictionary with tag: expression pairs of Conftool selectors.

        Yields:
            conftool.kvobject.Entity: the selected object.

        Raises:
            spicerack.confctl.ConfctlError: if not match is found.

        """
        selectors = {}
        for tag, expr in tags.items():
            selectors[tag] = re.compile('^{}$'.format(expr))

        obj = None
        for obj in self._entity.query(selectors):
            yield obj

        if obj is None:
            raise ConfctlError('No match found')

    def update(self, changed, **tags):
        """Updates the value of conftool objects corresponding to the selection done with tags.

        Arguments:
            changed (dict): the new values to set for the selected objects.
            **tags: arbitrary Conftool tags as keyword arguments.

        Raises:
            spicerack.confctl.ConfctlError: on etcd or Conftool errors.

        Examples:
            >>> confctl.update({'pooled': False}, service='appservers-.*', name='eqiad')

        """
        logger.debug('Updating conftool matching tags: %s', tags)
        if self._dry_run:
            message_prefix = 'Skipping conftool update on dry-run mode'
        else:
            message_prefix = 'Updating conftool'

        for obj in self._select(tags):
            logger.debug('%s: %s -> %s', message_prefix, obj, changed)
            if self._dry_run:
                continue

            try:
                obj.update(changed)
            except BackendError as e:
                raise ConfctlError('Error writing to etcd') from e
            except Exception as e:
                raise ConfctlError('Generic error in conftool') from e

    def get(self, **tags):
        """Generator that yields conftool objects corresponding to the selection.

        Arguments:
            **tags: arbitrary Conftool tags as keyword arguments.

        Yields:
            conftool.kvobject.Entity: the selected object.

        """
        for obj in self._select(tags):
            logger.debug('Selected conftool object: %s', obj)
            yield obj

    def set_and_verify(self, key, value, **tags):
        """Set and verify a single Conftool value.

        Arguments:
            key (str): the key in Conftool to modify.
            value (mixed): the value to set.
            **tags: arbitrary Conftool tags as keyword arguments.

        Raises:
            spicerack.confctl.ConfctlError: on etcd or Conftool errors or failing to verify the changes.

        """
        logger.info('Setting %s=%s for tags: %s', key, value, tags)
        self.update({key: value}, **tags)

        for obj in self.get(**tags):  # Verify the changes were applied
            new = getattr(obj, key)
            if new != value and not self._dry_run:
                raise ConfctlError("Conftool key {key} has value '{new}', expecting '{value}' for tags: {tags}".format(
                    key=key, new=new, value=value, tags=tags))



"""IPMI module tests."""
import os

from subprocess import CalledProcessError  # nosec
from unittest import mock

import pytest

from spicerack import ipmi


BOOTPARAMS_OUTPUT = """
Boot parameter version: 1
Boot parameter 5 is valid/unlocked
Boot parameter data: {bootparams}
 Boot Flags :
   - Boot Flag Invalid
   - Options apply to only next boot
   - BIOS PC Compatible (legacy) boot
   - Boot Device Selector : {pxe}
   - Console Redirection control : System Default
   - BIOS verbosity : Console redirection occurs per BIOS configuration setting (default)
   - BIOS Mux Control Override : BIOS uses recommended setting of the mux at the end of POST
Invalid line
"""


class TestIpmi:
    """Test class for the Ipmi class."""

    def setup_method(self):
        """Setup the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.ipmi = ipmi.Ipmi('password', dry_run=False)
        self.ipmi_dry_run = ipmi.Ipmi('password')

    def test_init(self):
        """It should initialize the instance and set the IPMITOOL_PASSWORD environment variable."""
        assert isinstance(self.ipmi, ipmi.Ipmi)
        assert os.getenv('IPMITOOL_PASSWORD') == 'password'

    @mock.patch('spicerack.ipmi.check_output', return_value=b'test')
    def test_command_ok(self, mocked_check_output):
        """It should execute the IPMI command as expected."""
        assert self.ipmi.command('test-mgmt.example.com', ['test_command']) == 'test'
        mocked_check_output.assert_called_once_with(
            ['ipmitool', '-I', 'lanplus', '-H', 'test-mgmt.example.com', '-U', 'root', '-E', 'test_command'])

    @mock.patch('spicerack.ipmi.check_output')
    def test_command_dry_run_ok(self, mocked_check_output):
        """It should not execute the IPMI command if in DRY RUN mode."""
        assert self.ipmi_dry_run.command('test-mgmt.example.com', ['test_command']) == ''
        assert not mocked_check_output.called

    @mock.patch('spicerack.ipmi.check_output')
    def test_command_raise(self, mocked_check_output):
        """It should raise IpmiError if failed to execute the command."""
        mocked_check_output.side_effect = CalledProcessError(1, 'executed_command')
        with pytest.raises(ipmi.IpmiError, match='Remote IPMI for test-mgmt.example.com failed'):
            self.ipmi.command('test-mgmt.example.com', ['test_command'])

    @mock.patch('spicerack.ipmi.check_output', return_value=b'Chassis Power is on')
    def test_check_connection_ok(self, mocked_check_output):
        """It should check that the connection to the remote IPMI works running a simple command."""
        self.ipmi_dry_run.check_connection('test-mgmt.example.com')
        mocked_check_output.assert_called_once_with(
            ['ipmitool', '-I', 'lanplus', '-H', 'test-mgmt.example.com', '-U', 'root', '-E',
             'chassis', 'power', 'status'])

    @mock.patch('spicerack.ipmi.check_output', return_value=b'failed')
    def test_check_connection_raise(self, mocked_check_output):
        """It should raise IpmiError if unable to execute remote IPMI commands."""
        with pytest.raises(ipmi.IpmiError, match='Unexpected chassis status: failed'):
            self.ipmi.check_connection('test-mgmt.example.com')

        assert mocked_check_output.called

    @mock.patch('spicerack.ipmi.check_output')
    def test_check_bootparams_ok(self, mocked_check_output):
        """It should check that the BIOS boot parameters are normal."""
        mocked_check_output.return_value = BOOTPARAMS_OUTPUT.format(bootparams='0000000000', pxe='No override').encode()
        self.ipmi.check_bootparams('test-mgmt.example.com')
        mocked_check_output.assert_called_once_with(
            ['ipmitool', '-I', 'lanplus', '-H', 'test-mgmt.example.com', '-U', 'root', '-E',
             'chassis', 'bootparam', 'get', '5'])

    @mock.patch('spicerack.ipmi.check_output')
    def test_check_bootparams_wrong_value(self, mocked_check_output):
        """It should raise IpmiCheckError if the BIOS boot parameters are not the normal ones."""
        mocked_check_output.return_value = BOOTPARAMS_OUTPUT.format(bootparams='0004000000', pxe='Force PXE').encode()
        with pytest.raises(ipmi.IpmiCheckError,
                           match=r"Expected BIOS boot params in \('0000000000', '8000020000'\) got: 0004000000"):
            self.ipmi_dry_run.check_bootparams('test-mgmt.example.com')

        assert mocked_check_output.called

    @mock.patch('spicerack.ipmi.check_output', return_value=b'Boot parameter data')
    def test_check_bootparams_unable_to_extract(self, mocked_check_output):
        """It should raise IpmiError if unable to extract the value of the BIOS boot parameters."""
        with pytest.raises(ipmi.IpmiError, match="Unable to extract value for parameter 'Boot parameter data'"):
            self.ipmi.check_bootparams('test-mgmt.example.com')

        assert mocked_check_output.called

    @mock.patch('spicerack.ipmi.check_output', return_value=b'Invalid')
    def test_check_bootparams_missing_label(self, mocked_check_output):
        """It should raise IpmiError if unable to find the label looked for."""
        with pytest.raises(ipmi.IpmiError, match="Unable to find the boot parameter 'Boot parameter data'"):
            self.ipmi.check_bootparams('test-mgmt.example.com')

        assert mocked_check_output.called

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    @mock.patch('spicerack.ipmi.check_output')
    def test_force_pxe_ok(self, mocked_check_output, mocked_sleep):
        """Should set the PXE boot mode for the next boot."""
        mocked_check_output.side_effect = [
            b'', BOOTPARAMS_OUTPUT.format(bootparams='0004000000', pxe='Force PXE').encode()]
        self.ipmi.force_pxe('test-mgmt.example.com')

        assert not mocked_sleep.called
        mocked_check_output.assert_has_calls([
            mock.call(['ipmitool', '-I', 'lanplus', '-H', 'test-mgmt.example.com', '-U', 'root', '-E',
                       'chassis', 'bootdev', 'pxe']),
            mock.call(['ipmitool', '-I', 'lanplus', '-H', 'test-mgmt.example.com', '-U', 'root', '-E',
                       'chassis', 'bootparam', 'get', '5'])])

    @mock.patch('spicerack.decorators.time.sleep', return_value=None)
    @mock.patch('spicerack.ipmi.check_output')
    def test_force_pxe_retried(self, mocked_check_output, mocked_sleep):
        """Should retry to set the PXE mode on failure."""
        mocked_check_output.side_effect = [
            b'PXE not set',
            BOOTPARAMS_OUTPUT.format(bootparams='0000000000', pxe='No override').encode(),
            b'PXE set',
            BOOTPARAMS_OUTPUT.format(bootparams='0004000000', pxe='Force PXE').encode()]
        self.ipmi.force_pxe('test-mgmt.example.com')
        assert mocked_sleep.called



"""Spicerack package."""
import os

from socket import gethostname

from pkg_resources import DistributionNotFound, get_distribution

from spicerack import interactive, puppet
from spicerack.administrative import Reason
from spicerack.confctl import Confctl
from spicerack.config import load_ini_config
from spicerack.debmonitor import Debmonitor
from spicerack.dns import Dns
from spicerack.dnsdisc import Discovery
from spicerack.elasticsearch_cluster import create_elasticsearch_clusters
from spicerack.icinga import Icinga, ICINGA_DOMAIN
from spicerack.ipmi import Ipmi
from spicerack.log import irc_logger
from spicerack.management import Management
from spicerack.mediawiki import MediaWiki
from spicerack.mysql import Mysql
from spicerack.phabricator import create_phabricator
from spicerack.redis_cluster import RedisCluster
from spicerack.remote import Remote


try:
    __version__ = get_distribution('wikimedia-spicerack').version  # Must be the same used as 'name' in setup.py
    """:py:class:`str`: the version of the current Spicerack module."""
except DistributionNotFound:  # pragma: no cover - this should never happen during tests
    pass  # package is not installed


class Spicerack:
    """Spicerack service locator."""

    def __init__(self, *, verbose=False, dry_run=True, cumin_config='/etc/cumin/config.yaml',
                 conftool_config='/etc/conftool/config.yaml', conftool_schema='/etc/conftool/schema.yaml',
                 debmonitor_config='/etc/debmonitor.conf',
                 spicerack_config_dir='/etc/spicerack'):
        """Initialize the service locator for the Spicerack library.

        Arguments:
            verbose (bool, optional): whether to set the verbose mode.
            dry_run (bool, optional): whether this is a DRY-RUN.
            cumin_config (str, optional): the path to Cumin's configuration file.
            conftool_config (str, optional): the path to Conftool's configuration file.
            conftool_schema (str, optional): the path to Conftool's schema file.
            debmonitor_config (str, optional): the path to Debmonitor's INI configuration file. It must have at least
                the following schema::

                    [DEFAULT]
                    server=debmonitor.example.com
                    cert=/etc/debmonitor/ssl/cert.pem
                    key=/etc/debmonitor/ssl/server.key

            spicerack_config_dir (str, optional): the path for the root configuration directory for Spicerack.
                Module-specific configuration will be loaded from `config_dir/module_name/`.
        """
        # Attributes
        self._verbose = verbose
        self._dry_run = dry_run
        self._cumin_config = cumin_config
        self._conftool_config = conftool_config
        self._conftool_schema = conftool_schema
        self._debmonitor_config = debmonitor_config
        self._spicerack_config_dir = spicerack_config_dir

        self._username = interactive.get_username()
        self._current_hostname = gethostname()
        self._irc_logger = irc_logger
        self._confctl = None

    @property
    def dry_run(self):
        """Getter for the ``dry_run`` property.

        Returns:
            bool: True if the DRY-RUN mode is set, False otherwise.

        """
        return self._dry_run

    @property
    def verbose(self):
        """Getter for the ``verbose`` property.

        Returns:
            bool: True if the verbose mode is set, False otherwise.

        """
        return self._verbose

    @property
    def username(self):
        """Getter for the current username.

        Returns:
            str: the name of the effective running user.

        """
        return self._username

    @property
    def irc_logger(self):
        """Getter for the ``irc_logger`` property.

        Returns:
            logging.Logger: the logger instance to write to IRC.

        """
        return self._irc_logger

    @property
    def icinga_master_host(self):
        """Getter for the ``icinga_master_host`` property.

        Returns:
            spicerack.remote.RemoteHosts: the instance to execute commands on the Icinga master host.

        """
        return self.remote().query(self.dns().resolve_cname(ICINGA_DOMAIN))

    def remote(self):
        """Get a Remote instance.

        Returns:
            spicerack.remote.Remote: the Remote instance.

        """
        return Remote(self._cumin_config, dry_run=self._dry_run)

    def confctl(self, entity_name):
        """Access a Conftool specific entity instance.

        Arguments:
            entity_name (str): the name of a Conftool entity. Possible values: ``node``, ``service``, ``discovery``,
                ``mwconfig``.

        Returns:
            spicerack.confctl.ConftoolEntity: the confctl entity instance.

        """
        if self._confctl is None:
            self._confctl = Confctl(config=self._conftool_config, schema=self._conftool_schema, dry_run=self._dry_run)

        return self._confctl.entity(entity_name)

    def dns(self):  # pylint: disable=no-self-use
        """Get a Dns instance.

        Returns:
            spicerack.dns.Dns: a Dns instance that will use the operating system default namserver(s).

        """
        return Dns()

    def discovery(self, *records):
        """Get a Discovery instance.

        Arguments:
            *records (str): arbitrary positional arguments, each one must be a Discovery DNS record name.

        Returns:
            spicerack.dnsdisc.Discovery: the pre-configured Discovery instance for the given records.

        """
        return Discovery(self.confctl('discovery'), self.remote(), records, dry_run=self._dry_run)

    def mediawiki(self):
        """Get a MediaWiki instance.

        Returns:
            spicerack.mediawiki.MediaWiki: the MediaWiki instance.

        """
        return MediaWiki(self.confctl('mwconfig'), self.remote(), self._username, dry_run=self._dry_run)

    def mysql(self):
        """Get a Mysql instance.

        Returns:
            spicerack.mysql.Mysql: the Mysql instance.

        """
        return Mysql(self.remote(), dry_run=self._dry_run)

    def redis_cluster(self, cluster):
        """Get a RedisCluster instance.

        Arguments:
            cluster (str): the name of the cluster.

        Returns:
            spicerack.redis_cluster.RedisCluster: the cluster instance.

        """
        return RedisCluster(cluster, os.path.join(self._spicerack_config_dir, 'redis_cluster'), dry_run=self._dry_run)

    def elasticsearch_clusters(self, clustergroup):
        """Get an ElasticsearchClusters instance.

        Arguments:
            clustergroup (str): name of cluster group e.g ``search_eqiad``.

        Returns:
            spicerack.elasticsearch_cluster.ElasticsearchClusters: ElasticsearchClusters instance.

        """
        return create_elasticsearch_clusters(clustergroup, self.remote(), dry_run=self._dry_run)

    def admin_reason(self, reason, task_id=None):
        """Get an administrative Reason instance.

        Arguments:
            reason (str): the reason to use to justify an administrative action. See `spicerack.administrative.Reason`
                for all the details.
            task_id (str, optional): the task ID to mention in the reason.

        Returns:
            spicerack.administrative.Reason: the administrative Reason instance.

        """
        return Reason(reason, self._username, self._current_hostname, task_id=task_id)

    def icinga(self):
        """Get an Icinga instance.

        Returns:
            spicerack.icinga.Icinga: Icinga instance.

        """
        return Icinga(self.icinga_master_host)

    def puppet(self, remote_hosts):  # pylint: disable=no-self-use
        """Get a PuppetHosts instance for the given remote hosts.

        Arguments:
            remote_hosts (spicerack.remote.RemoteHosts): the instance with the target hosts.

        Returns:
            spicerack.puppet.PuppetHosts: the instance to manage Puppet on the target hosts.

        """
        return puppet.PuppetHosts(remote_hosts)

    def puppet_master(self):
        """Get a PuppetMaster instance to manage hosts and certificates from a Puppet master.

        Returns:
            spicerack.puppet.PuppetMaster: the instance to manage Puppet hosts and certificates.

        """
        return puppet.PuppetMaster(self.remote().query(puppet.get_puppet_ca_hostname()))

    def ipmi(self):
        """Get an Ipmi instance to send remote IPMI commands to management consoles.

        Returns:
            spicerack.ipmi.Ipmi: the instance to run ipmitool commands.

        """
        return Ipmi(interactive.get_management_password(), dry_run=self._dry_run)

    def phabricator(self, bot_config_file, section='phabricator_bot'):
        """Get a Phabricator instance to interact with a Phabricator website.

        Arguments:
            bot_config_file (str): the path to the configuration file for the Phabricator bot, with the following
                structure::

                    [section_name]
                    host = https://phabricator.example.org/api/
                    username = phab-bot
                    token = api-12345

            section (str, optional): the name of the section of the configuration file where to find the required
                parameters.

        Returns:
            spicerack.phabricator.Phabricator: the instance.

        """
        # Allow to specify the configuration file as opposed to other methods so that different clients can use
        # different Phabricator BOT accounts, potentially with different permissions.
        return create_phabricator(bot_config_file, section=section, dry_run=self._dry_run)

    def debmonitor(self):
        """Get a Debmonitor instance to interact with a Debmonitor website.

        Returns:
            spicerack.debmonitor.Debmonitor: the instance.

        Raises:
            KeyError: if any configuration option is missing.

        """
        options = load_ini_config(self._debmonitor_config).defaults()
        return Debmonitor(options['server'], options['cert'], options['key'], dry_run=self._dry_run)

    def management(self):
        """Get a Management instance to interact with the management interfaces.

        Returns:
            spicerack.management.Management: the instance.

        """
        return Management(self.dns())



"""Tests package for Spicerack."""
import os

from pkg_resources import get_distribution, parse_version

import pytest


TESTS_BASE_PATH = os.path.realpath(os.path.dirname(__file__))
CAPLOG_MIN_VERSION = '3.3.0'
REQUESTS_MOCK_MIN_VERSION = '1.5.0'
ELASTICSEARCH_MIN_VERSION = '5.0.0'


def get_fixture_path(*paths):
    """Return the absolute path of the given fixture.

    Arguments:
        *paths: arbitrary positional arguments used to compose the absolute path to the fixture.

    Returns:
        str: the absolute path of the selected fixture.

    """
    return os.path.join(TESTS_BASE_PATH, 'fixtures', *paths)


SPICERACK_TEST_PARAMS = {
    'cumin_config': get_fixture_path('remote', 'config.yaml'),
    'conftool_config': get_fixture_path('confctl', 'config.yaml'),
    'conftool_schema': get_fixture_path('confctl', 'schema.yaml'),
    'debmonitor_config': get_fixture_path('debmonitor', 'config.ini'),
    'spicerack_config_dir': get_fixture_path(),
}


def caplog_not_available():
    """Check if the caplog fixture is not available.

    Returns:
        bool: True if the caplog fixture is not available, False otherwise.

    """
    return parse_version(pytest.__version__) < parse_version(CAPLOG_MIN_VERSION)


def requests_mock_not_available():
    """Check if the requests_mock fixture is not available.

    Returns:
        bool: True if the requests_mock fixture is not available, False otherwise.

    """
    return parse_version(get_distribution('requests_mock').version) < parse_version(REQUESTS_MOCK_MIN_VERSION)


def elasticsearch_too_old():
    """Check if elasticsearch version is less than 5.0.0

    Returns:
        bool: True if elasticsearch version is less 5.0.0, False if otherwise

    """
    return parse_version(get_distribution('elasticsearch').version) < parse_version(ELASTICSEARCH_MIN_VERSION)



"""Remote module to execute commands on hosts via Cumin."""
import logging

from datetime import datetime, timedelta

from cumin import Config, CuminError, query, transport, transports
from cumin.cli import target_batch_size

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class RemoteError(SpicerackError):
    """Custom exception class for errors of this module."""


class RemoteCheckError(SpicerackCheckError):
    """Custom exception class for check errors of this module."""


class RemoteExecutionError(RemoteError):
    """Custom exception class for remote execution errors."""

    def __init__(self, retcode, message):
        """Override parent constructor to add the return code attribute."""
        super().__init__('{msg} (exit_code={ret})'.format(msg=message, ret=retcode))
        self.retcode = retcode


class RemoteHostsAdapter:
    """Base adapter to write classes that expand the capabilities of RemoteHosts.

    This adapter class is a helper class to reduce duplication when writing classes that needs to add capabilities to a
    RemoteHosts instance. The goal is to not extend the RemoteHosts but instead delegate to its instances.
    This class fits when a single RemoteHosts instance is enough, but for more complex cases, in which multiple
    RemoteHosts instances should be orchestrated, it's ok to not extend this class and create a standalone one.
    """

    def __init__(self, remote_hosts):
        """Initialize the instance.

        Arguments:
            remote_hosts (spicerack.remote.RemoteHosts): the instance to act on the remote hosts.
        """
        self._remote_hosts = remote_hosts

    def __str__(self):
        """String representation of the instance.

        Returns:
            str: the string representation of the target hosts.

        """
        return str(self._remote_hosts)

    def __len__(self):
        """Length of the instance.

        Returns:
            int: the number of target hosts.

        """
        return len(self._remote_hosts)


class Remote:
    """Remote class to interact with Cumin."""

    def __init__(self, config, dry_run=True):
        """Initialize the instance.

        Arguments:
            config (str): the path of Cumin's configuration file.
            dry_run (bool, optional): whether this is a DRY-RUN.
        """
        self._config = Config(config)
        self._dry_run = dry_run

    def query(self, query_string):
        """Execute a Cumin query and return the matching hosts.

        Arguments:
            query_string (str): the Cumin query string to execute.

        Returns:
            spicerack.remote.RemoteHosts: RemoteHosts instance matching the given query.

        """
        try:
            hosts = query.Query(self._config).execute(query_string)
        except CuminError as e:
            raise RemoteError('Failed to execute Cumin query') from e

        return RemoteHosts(self._config, hosts, dry_run=self._dry_run)


class RemoteHosts:
    """Remote Executor class.

    This class can be extended to customize the interaction with remote hosts passing a custom factory function to
    `spicerack.remote.Remote.query`.
    """

    def __init__(self, config, hosts, dry_run=True):
        """Initialize the instance.

        Arguments:
            config (cumin.Config): the configuration for Cumin.
            hosts (ClusterShell.NodeSet.NodeSet): the hosts to target for the remote execution.
            dry_run (bool, optional): whether this is a DRY-RUN.

        Raises:
            spicerack.remote.RemoteError: if no hosts were provided.

        """
        if not hosts:
            raise RemoteError('No hosts provided')

        self._config = config
        self._hosts = hosts
        self._dry_run = dry_run

    @property
    def hosts(self):
        """Getter for the hosts property.

        Returns:
            ClusterShell.NodeSet.NodeSet: a copy of the targeted hosts.

        """
        return self._hosts.copy()

    def __str__(self):
        """String representation of the instance.

        Returns:
            str: the string representation of the target hosts.

        """
        return str(self._hosts)

    def __len__(self):
        """Length of the instance.

        Returns:
            int: the number of target hosts.

        """
        return len(self._hosts)

    def run_async(self, *commands, success_threshold=1.0, batch_size=None, batch_sleep=None, is_safe=False):
        """Execute commands on hosts matching a query via Cumin in async mode.

        Arguments:
            *commands (str, cumin.transports.Command): arbitrary number of commands to execute on the target hosts.
            success_threshold (float, optional): to consider the execution successful, must be between 0.0 and 1.0.
            batch_size (int, str, optional): the batch size for cumin, either as percentage (e.g. ``25%``)
                or absolute number (e.g. ``5``).
            batch_sleep (float, optional): the batch sleep in seconds to use in Cumin before scheduling the next host.
            is_safe (bool, optional): whether the command is safe to run also in dry-run mode because it's a read-only
                command that doesn't modify the state.

        Returns:
            generator: cumin.transports.BaseWorker.get_results to allow to iterate over the results.

        Raises:
            RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        return self._execute(list(commands), mode='async', success_threshold=success_threshold, batch_size=batch_size,
                             batch_sleep=batch_sleep, is_safe=is_safe)

    def run_sync(self, *commands, success_threshold=1.0, batch_size=None, batch_sleep=None, is_safe=False):
        """Execute commands on hosts matching a query via Cumin in sync mode.

        Arguments:
            *commands (str, cumin.transports.Command): arbitrary number of commands to execute on the target hosts.
            success_threshold (float, optional): to consider the execution successful, must be between 0.0 and 1.0.
            batch_size (int, str, optional): the batch size for cumin, either as percentage (e.g. ``25%``)
                or absolute number (e.g. ``5``).
            batch_sleep (float, optional): the batch sleep in seconds to use in Cumin before scheduling the next host.
            is_safe (bool, optional): whether the command is safe to run also in dry-run mode because it's a read-only
                command that doesn't modify the state.

        Returns:
            generator: cumin.transports.BaseWorker.get_results to allow to iterate over the results.

        Raises:
            RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        return self._execute(list(commands), mode='sync', success_threshold=success_threshold, batch_size=batch_size,
                             batch_sleep=batch_sleep, is_safe=is_safe)

    def reboot(self, batch_size=1, batch_sleep=180.0):
        """Reboot hosts.

        Arguments:
            batch_size (int, optional): how many hosts to reboot in parallel.
            batch_sleep (float, optional): how long to sleep between one reboot and the next.
        """
        if len(self._hosts) == 1:  # Temporary workaround until T213296 is fixed.
            batch_sleep = None

        logger.info('Rebooting %d hosts in batches of %d with %.1fs of sleep in between: %s',
                    len(self._hosts),
                    batch_size,
                    batch_sleep if batch_sleep is not None else 0.0,
                    self._hosts)

        self.run_sync(transports.Command('reboot-host', timeout=30), batch_size=batch_size, batch_sleep=batch_sleep)

    @retry(tries=25, delay=timedelta(seconds=10), backoff_mode='linear',
           exceptions=(RemoteExecutionError, RemoteCheckError))
    def wait_reboot_since(self, since):
        """Poll the host until is reachable and has an uptime lower than the provided datetime.

        Arguments:
            since (datetime.datetime): the time after which the host should have booted.

        Raises:
            spicerack.remote.RemoteCheckError: if unable to connect to the host or the uptime is higher than expected.

        """
        remaining = self.hosts
        delta = (datetime.utcnow() - since).total_seconds()
        for nodeset, uptime in self.uptime():
            if uptime >= delta:
                raise RemoteCheckError('Uptime for {hosts} higher than threshold: {uptime} > {delta}'.format(
                    hosts=nodeset, uptime=uptime, delta=delta))

            remaining.difference_update(nodeset)

        if remaining:
            raise RemoteCheckError('Unable to check uptime from {num} hosts: {hosts}'.format(
                num=len(remaining), hosts=remaining))

        logger.info('Found reboot since %s for hosts %s', since, self._hosts)

    def uptime(self):
        """Get current uptime.

        Returns:
            list: a list of 2-element :py:class:`tuple` instances with hosts :py:class:`ClusterShell.NodeSet.NodeSet`
            as first item and :py:class:`float` uptime as second item.

        """
        results = self.run_sync(transports.Command('cat /proc/uptime', timeout=10), is_safe=True)
        # Callback to extract the uptime from /proc/uptime (i.e. getting 12345.67 from '12345.67 123456789.00').
        return RemoteHosts.results_to_list(results, callback=lambda output: float(output.split()[0]))

    def init_system(self):
        """Detect the init system.

        Returns:
            list: a list of 2-element tuples with hosts :py:class:`ClusterShell.NodeSet.NodeSet` as first item and the
            init system :py:class:`str` as second.

        """
        results = self.run_sync(transports.Command('ps --no-headers -o comm 1', timeout=10), is_safe=True)
        return RemoteHosts.results_to_list(results)

    @staticmethod
    def results_to_list(results, callback=None):
        """Extract execution results into a list converting them with an optional callback.

        Todo:
            move it directly into Cumin.

        Arguments:
            results (generator): generator returned by run_sync() and run_async() to iterate over the results.
            callback (callable, optional): an optional callable to apply to each result output (it can be multiline).
                The callback will be called with a the string output as the only parameter and must return the
                extracted value. The return type can be chosen freely.

        Returns:
            list: a list of 2-element tuples with hosts :py:class:`ClusterShell.NodeSet.NodeSet` as first item and the
            extracted outputs :py:class:`str` as second. This is because NodeSet are not hashable.

        Raises:
            spicerack.remote.RemoteError: if unable to run the callback.

        """
        extracted = []
        for nodeset, output in results:
            result = output.message().decode().strip()
            if callback is not None:
                try:
                    result = callback(result)
                except Exception as e:
                    raise RemoteError('Unable to extract data with {cb} for {hosts} from: {output}'.format(
                        cb=callback.__name__, hosts=nodeset, output=result)) from e

            extracted.append((nodeset, result))

        return extracted

    def _execute(self, commands, mode='sync', success_threshold=1.0,  # pylint: disable=too-many-arguments
                 batch_size=None, batch_sleep=None, is_safe=False):
        """Lower level Cumin's execution of commands on the target nodes.

        Arguments:
            commands (list): the list of commands to execute on the target hosts, either a list of commands or a list
                of cumin.transports.Command instances.
            mode (str, optional): the Cumin's mode of execution. Accepted values: sync, async.
            success_threshold (float, optional): to consider the execution successful, must be between 0.0 and 1.0.
            batch_size (int, str, optional): the batch size for cumin, either as percentage (e.g. ``25%``) or absolute
                number (e.g. ``5``).
            batch_sleep (float, optional): the batch sleep in seconds to use in Cumin before scheduling the next host.
            is_safe (bool, optional): whether the command is safe to run also in dry-run mode because it's a read-only
                command that doesn't modify the state.

        Returns:
            generator: as returned by :py:meth:`cumin.transports.BaseWorker.get_results` to iterate over the results.

        Raises:
            RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        if batch_size is None:
            parsed_batch_size = {'value': None, 'ratio': None}
        else:
            parsed_batch_size = target_batch_size(str(batch_size))

        target = transports.Target(
            self._hosts, batch_size=parsed_batch_size['value'], batch_size_ratio=parsed_batch_size['ratio'],
            batch_sleep=batch_sleep)
        worker = transport.Transport.new(self._config, target)
        worker.commands = commands
        worker.handler = mode
        worker.success_threshold = success_threshold

        logger.debug('Executing commands %s on %d hosts: %s', commands, len(target.hosts), str(target.hosts))

        if self._dry_run and not is_safe:
            return iter(())  # Empty generator

        ret = worker.execute()

        if ret != 0 and not self._dry_run:
            raise RemoteExecutionError(ret, 'Cumin execution failed')

        return worker.get_results()
